#include <ncurses.h>
#include <clocale>
#include "string"
#include <filesystem>
#include <fstream>
#include <vector>
#include <memory>
#include <algorithm>
#include <set>
#include <cstring>
#include <cstdio>

#include "Disc.h"

bool thread_run = true;

using DiscList=std::vector<std::unique_ptr<Disc>>;

// gui over www

static constexpr std::chrono::seconds disc_pooling_time{3};
void scanForDiscs(DiscList& disc_list)
{
    std::set<std::tuple<std::filesystem::path, std::string>> sys_list;

    for(const auto& p: std::filesystem::directory_iterator("/sys/block"))
    {
        std::string wwid;
        std::ifstream(p.path()/"device/wwid") >> wwid;
        sys_list.insert({p, wwid});
    }

    for(const auto& p: sys_list)
    {
        unsigned long dev_size;
        std::ifstream(std::get<0>(p)/"size") >> dev_size;
        std::string filename{std::get<0>(p).filename().string()};
        auto dev_path = std::filesystem::path{"/dev"}/filename;

        auto disc = std::find_if(disc_list.begin(), disc_list.end(), [&dev_path, &p](const std::unique_ptr<Disc>& a)
                             {return a->getPath() == dev_path && a->getWwid() == std::get<1>(p);});
        if(disc != disc_list.end())
        {
            continue;
        }

        dev_size <<=9;
        std::string mount_path;
        std::ifstream mounts("/proc/mounts");
        bool mounted = false;
        while (mounts >> mount_path)
        {
            if(mount_path.size() >= filename.size() + 5 &&
                    mount_path.compare(5, filename.size(), filename) == 0)
            {
                mounted = true;
                break;
            }
        }

        if(!mounted && dev_size > 0)
        {
            bool have_name = false;
            if (std::filesystem::exists("paths.conf"))
            {
                std::ifstream paths("paths.conf");
                std::string line;
                while (paths >> line)
                {
                    auto path = std::filesystem::path{"/dev/disk/by-path"}/line;
                    if(paths >> line
                            && std::filesystem::exists(path)
                            && std::filesystem::is_symlink(path)
                            &&  std::filesystem::canonical(
                                    std::filesystem::path{"/dev/disk/by-path"}/std::filesystem::read_symlink(path)) == dev_path)
                    {
                        filename = line;
                        have_name = true;
                        break;
                    }
                }
            }

            std::array<char, 25> model{};
            std::ifstream(std::get<0>(p)/"device/model").getline(model.data(), model.size());
            disc_list.emplace_back( new Disc(dev_path, have_name
                                  ? filename : dev_path.string(), std::string{model.data()}, std::get<1>(p), dev_size));
        }
    }

    for (auto it = disc_list.begin(); it != disc_list.end(); ) {
        auto p = std::find_if(sys_list.begin(), sys_list.end(), [&it](const std::tuple<std::filesystem::path, std::string>& a)
                              {return std::filesystem::path{"/dev"}/std::get<0>(a).filename() == it->get()->getPath()
                                      && it->get()->getWwid() == std::get<1>(a);});
        if(p == sys_list.end())
        {
            disc_list.erase(it);
        }
        else
        {
            ++it;
        }
    }

    std::sort(disc_list.begin(), disc_list.end(), [](std::unique_ptr<Disc>& a, std::unique_ptr<Disc>& b) {
        return a->getName() < b->getName();
    });
}

int main()
{
    enum Colors{
        WHITE = 2,
        YELLOW = 3,
        RED = 4,
        GREEN = 5
    };
    int rows, columns;
    setlocale(LC_ALL, "en_US.UTF-8");
    initscr();
    start_color();
    cbreak();
    noecho();
    keypad(stdscr, true);
    getmaxyx(stdscr,rows,columns);
    curs_set(0);
    init_pair(1, COLOR_BLACK, COLOR_YELLOW);
    init_pair(WHITE, COLOR_WHITE, COLOR_BLACK);
    init_pair(YELLOW, COLOR_YELLOW, COLOR_BLACK);
    init_pair(RED, COLOR_RED, COLOR_BLACK);
    init_pair(GREEN, COLOR_GREEN, COLOR_BLACK);

    addstr( " DWIPE");
    mvchgat(0,0, -1, A_NORMAL, 1, NULL);
    refresh();

    WINDOW* menu = newwin(1, 120, rows-2, 2);
    const auto menu_str = "ENTER: Start zeroing | BACKSPACE: Abort | R: Restart the application | Q: Shutdown the PC";
    wprintw( menu, menu_str);
    wrefresh(menu);

    halfdelay(2);
    unsigned int a;
    DiscList dist_list;
    scanForDiscs(dist_list);

    WINDOW* selector = newwin(rows-4, 1, 2, 2);
    WINDOW* window = newwin(rows-4, 120, 2, 4);
    keypad(window, true);

    int disc_selector = 0;
    auto start = std::chrono::high_resolution_clock::now();
    while (thread_run)
    {
        auto duration{std::chrono::high_resolution_clock::now() - start};
        if (duration > disc_pooling_time)
        {
            scanForDiscs(dist_list);
            start = std::chrono::high_resolution_clock::now();
        }

        int index = 0;
        werase(window);
        for(const auto& disc: dist_list)
        {
            float done = (float)disc->getZeroWriter().getWriten()/(float)disc->getSize();
            mvwprintw(window, index, 0, "● %-15s ", disc->getName().c_str());
            for (int i = done / 0.003125; i >= 0; i-=8)
            {
                if(i<8)
                {
                    wprintw(window, "%lc", L'▏' - i);
                }
                else
                {
                    waddch(window, ' ' | A_REVERSE);
                }
            }
            mvwprintw(window, index, 58, "▏ %3.2f%%  %.2fMB/s %luMB/%luMB", done*100,
                    (float)disc->getZeroWriter().getSpeed()/1000000,
                    disc->getZeroWriter().getWriten()/1000000,
                    disc->getZeroWriter().getSize()/1000000);

            /* Print model */
            mvwprintw(window, index+1, 2, "%s",
                    disc->getModel().c_str());

            switch (disc->getState()) {
            case Disc::State::CHECKING_BUFFER_SIZE:
                mvwprintw(window, index+2, 2, "%s [buffer %u]",
                        Disc::state_strings[disc->getState()],
                        disc->getBufferSizeChecker().getCurrentBufferSize());
                break;
            case Disc::State::ERROR:
                mvwprintw(window, index+2, 2, "%s %s",
                        Disc::state_strings[disc->getState()],
                        std::strerror(disc->getLastError()));
                break;
            default:
                mvwprintw(window, index+2, 2, "%s", Disc::state_strings[disc->getState()]);
                break;
            }

            switch (disc->getState()) {
            case Disc::State::IDLE:
                mvwchgat(window, index,0, 2, A_NORMAL, Colors::WHITE, NULL);
                break;
            case Disc::State::ERROR:
                mvwchgat(window, index,0, 2, A_NORMAL, Colors::RED, NULL);
                break;
            case Disc::State::ABORTING:
            case Disc::State::END_OF_CONTROL_STATES:
            case Disc::State::CHECKING_HEALTH:
            case Disc::State::CHECKING_BUFFER_SIZE:
            case Disc::State::ZEROING:
                mvwchgat(window, index,0, 2, A_NORMAL, Colors::YELLOW, NULL);
                break;
            case Disc::State::DONE_PING:
                mvwchgat(window, index,0, 2, A_NORMAL, Colors::GREEN, NULL);
                break;
            }
            mvwchgat(window, index,2, 15, A_BOLD, 0, NULL);

            index += 3;
        }

        mvwaddch(selector, (disc_selector*3),0,'>');
        wrefresh(selector);
        wrefresh(menu);
        refresh();

        a = wgetch(window);
        switch (a) {
        case 'q':
            werase(menu);
            wprintw( menu, "Y: to shutdown");
            wrefresh(menu);
            halfdelay(20);
            a = wgetch(menu);
            if(a == 'y'){
                popen("poweroff -p","r");
            }
            halfdelay(2);
            werase(menu);
            wprintw( menu, menu_str);
            wrefresh(menu);
            break;
        case 'r':
            werase(menu);
            wprintw( menu, "Y: to restart");
            wrefresh(menu);
            halfdelay(20);
            a = wgetch(menu);
            if(a == 'y'){
                thread_run=false;
            }
            halfdelay(2);
            werase(menu);
            wprintw( menu, menu_str);
            wrefresh(menu);
            break;
        case 127:
        case 0407:
        case KEY_BACKSPACE:
            if((std::vector<Disc>::size_type)disc_selector < dist_list.size())
            {
                dist_list[disc_selector]->abort();
            }
            break;
        case 10:
        case 0527:
        case KEY_ENTER:
            if((std::vector<Disc>::size_type)disc_selector < dist_list.size())
            {
                dist_list[disc_selector]->startZeroing();
            }
            break;
        case 'w':
        case 259:
            if(disc_selector > 0)
            {
                werase(selector);
                disc_selector--;
            }
            break;
        case 's':
        case 258:
            if((std::vector<Disc>::size_type)disc_selector < dist_list.size()-1)
            {
                werase(selector);
                disc_selector++;
            }
            break;
        default:
            break;
        }
    }

    endwin();
    return 0;
}
